// ----------------------------------------------------------------------------------
// Final exercise for the C++ Programming Course, 
// Dept. of Informatics, Athens Univ. of Economics and Business
//
// May 2024 Examination Period
//
// G. Papaioannou

#include "generator.h"
#include <iostream>
#include <vector>
#include <string>
#include <algorithm>

// Do not touch the following 4 lines
#ifdef EXAMINER_CODE
	#include "test.h"
	#include <assert.h>
#endif

//--------------------------------------------------------------------------------------------------
//
// Unit Tests.
//
// Instructions: Successively uncomment the following defines to enable the tests one by one and 
//               progressively add your functionality to the SocialNode class
// See the exam. paper for the grading breakdown
//
// General hints:
// - Uncomment the tests starting from TEST1 and gradually implement all required functionality
// - Be aware of function/method argument requirements. Should I pass the argument by reference?
// - You may benefit from choosing the appropriate internal data representation for SocialNode members.

//#define TEST1
//#define TEST2
//#define TEST3
//#define TEST4
//#define TEST5

//--------------------------------------------------------------------------------------------------

#ifdef TEST1

#include "social.h"
//
// Forward fuction declarations
//
// TODO: Declare a class SocialNode as a subclass of Node<T> 
//       with template argument std::string (see declaration in node.h)
std::vector<SocialNode*> generateNetwork(int num_nodes);

void listUserFriends(const SocialNode& node);
#endif

//--------------------------------------------------------------------------------------------------


int main(int argc, char ** argv)
{
	SRAND(0);
	size_t uuid = RND();
	int N_users = 10 + uuid % 20;

	std::vector<SocialNode*> social_network;

#ifdef TEST2
	//
	// Generate a small social network:
	// (see generateNetwork function definition below)
	//
	social_network = generateNetwork(N_users);
#endif

#ifdef TEST3
	//
	// List the friends of a random user:
	// (see listUserFriends fuction definition below)
	//
	listUserFriends(*social_network[uuid % N_users]);
#endif

#ifdef TEST4
	// Ask whether two arbitrary users are either directly or indirectly connected
	SocialNode * user1 = social_network[RND() % social_network.size()];
	SocialNode * user2 = social_network[RND() % social_network.size()];
	
	// TODO: declare and define the missing method:
	//       should return true if the SocialNode parameter 
	//       is connected to the calling instance either directly or via any number of indirect connections.
	//       Returns false otherwise.
	//       For example, the nodes A,B,C,D,E are connected as follows: A<-->B, B<-->C, D<-->E.
	//       A is "associated" to C, but not to D.
	//       Hints: a) the method should recursively explore the friends of the friends (via iterative implementation).
	//       b) the method should keep track of visited nodes during this exploration, to avoid running indefinately in circles
	bool result = user1->associated(*user2);

	std::cout << "Is user " << user1->getUserName() << " connected to " << user2->getUserName() << "? ";
	std::cout << (result ? "YES\n\n" : "NO\n\n");
#endif

	// Do not touch the following 3 lines
	#if defined (EXAMINER_CODE) && defined (TEST1) && defined (TEST2)
		assert(testNetworkIntegrity(social_network));
	#endif

#ifdef TEST5
	// Discard the network before quiting
	// TODO: properly declare and implement the following function in social.h to erase the network, 
	//       in order to avoid memory leaks.
	deleteSocialNetwork(social_network);
#endif

	printf("\nPress ENTER to continue...");
	getchar();
}

//------------------------------------------------------------------------------

#ifdef TEST2
std::vector<SocialNode*> generateNetwork(int num_nodes)
{
	std::vector<SocialNode*> nds;

	// Make a population of num_nodes users of the social network
	for (int i = 0; i < num_nodes; i++)
	{
		// TODO: declare and define the appropriate constructor:
		nds.push_back(new SocialNode(generateName()));
	}

	for (int i = 0; i < num_nodes; i++)
	{
		// establish up to 10 social network connections with other users
		
		// befriend (connect to) a random node
		// TODO: declare and define the appropriate operator:
		// IMPORTANT: 
		// a) The friends of a user must be UNIQUE: 
		//    no duplicate connections are allowed.
		// b) When user A connects to user B, the inverse connection is 
		//    also established: B connects to A:
		//    When A becomes a friend of B, B also becomes a friend of A
		*nds[i] += nds[RND() % num_nodes];

		for (unsigned int j = 1; j <= 5 + RND() % 5; j++)
		{
			if (RAND0TO1() > 0.2f)
				// befriend (connect to) a random node with a probability of 80%
			{
				*nds[i] += nds[RND() % num_nodes];
			}
		}
	}
	return nds;
}
#endif

//------------------------------------------------------------------------------

#ifdef TEST3
void listUserFriends(const SocialNode& node)
{
	// TODO: declare and define the appropriate SocialNode methods
	if (node.getNumFriends() == 0)
	{
		std::cout << "User " + node.getUserName() + " has no friends.\n\n";
		return;
	}

	std::cout << "User " + node.getUserName() + " has the following " << node.getNumFriends() << " friends:\n";
	std::vector<SocialNode*> friends = node.getFriends();
	
	for (auto user : friends)
		std::cout << "\t" << user->getUserName() << "\n";
	std::cout << "\n";

}
#endif


